
import 'dart:math' as math;

import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/painting.dart';
import 'package:nui_core/nui_core.dart';

enum NUIBannerLocation {
  topStart,
  topEnd,
  bottomStart,
  bottomEnd,
}

class BannerPainter extends CustomPainter {
  BannerPainter({
    @required this.message,
    @required this.textDirection,
    @required this.location,
    @required this.layoutDirection,
    this.color: const Color(0xA0B71C1C),
    this.height: 12,
    this.textStyle,
    this.gradient,
    this.shadow : const BoxShadow(
      color: Colors.black12,
      blurRadius: 6.0,
    )
  }) : assert(message != null),
        assert(textDirection != null),
        assert(location != null),
        assert(color != null),
        assert(textStyle != null),
        super(repaint: PaintingBinding.instance.systemFonts){
    this.height = this.height ?? 12;
    this.textStyle = this.textStyle ?? TextStyle(
      color: Color(0xFFFFFFFF),
      fontSize: height * 0.85,
      fontWeight: FontWeight.w900,
      height: 1.0,
    );
    this.color = this.color ?? Color(0xA0B71C1C);
  }

  double _kOffset = 40.0; // distance to bottom of banner, at a 45 degree angle inwards
  double get _kBottomOffset => _kOffset + 0.707 * this.height; // offset plus sqrt(2)/2 * banner height
  Rect get _kRect => Rect.fromLTWH(-_kOffset, _kOffset - this.height, _kOffset * 2.0, this.height);

  double height;
  final String message;
  final TextDirection textDirection;
  final NUIBannerLocation location;
  final TextDirection layoutDirection;
  Color color;
  final LinearGradient gradient;
  TextStyle textStyle;
  final BoxShadow shadow;

  bool _prepared = false;
  TextPainter _textPainter;
  Paint _paintShadow;
  Paint _paintBanner;

  void _prepare() {
    _paintShadow = shadow.toPaint();

    if(gradient != null) {
      _paintBanner = Paint()
        ..shader = gradient.createShader(_kRect);
    }
    else{
      _paintBanner = Paint()
        ..color = color;
    }

    _textPainter = TextPainter(
      text: TextSpan(style: textStyle, text: message),
      textAlign: TextAlign.center,
      textDirection: textDirection,
    );
    _prepared = true;
  }

  @override
  void paint(Canvas canvas, Size size) {
    if (!_prepared)
      _prepare();
    canvas
      ..translate(_translationX(size.width), _translationY(size.height))
      ..rotate(_rotation)
      ..drawRect(_kRect, _paintShadow)
      ..drawRect(_kRect, _paintBanner);
    double width = _kOffset * 2.0;
    _textPainter.layout(minWidth: width, maxWidth: width);
    _textPainter.paint(canvas, _kRect.topLeft + Offset(0.0, (_kRect.height - _textPainter.height) / 2.0));
  }

  @override
  bool shouldRepaint(BannerPainter oldDelegate) {
    return message != oldDelegate.message
        || location != oldDelegate.location
        || color != oldDelegate.color
        || textStyle != oldDelegate.textStyle;
  }

  @override
  bool hitTest(Offset position) => false;

  double _translationX(double width) {
    assert(location != null);
    assert(layoutDirection != null);
    switch (layoutDirection) {
      case TextDirection.rtl:
        switch (location) {
          case NUIBannerLocation.bottomEnd:
            return _kBottomOffset;
          case NUIBannerLocation.topEnd:
            return 0.0;
          case NUIBannerLocation.bottomStart:
            return width - _kBottomOffset;
          case NUIBannerLocation.topStart:
            return width;
        }
        break;
      case TextDirection.ltr:
        switch (location) {
          case NUIBannerLocation.bottomEnd:
            return width - _kBottomOffset;
          case NUIBannerLocation.topEnd:
            return width;
          case NUIBannerLocation.bottomStart:
            return _kBottomOffset;
          case NUIBannerLocation.topStart:
            return 0.0;
        }
        break;
    }
    return null;
  }

  double _translationY(double height) {
    assert(location != null);
    switch (location) {
      case NUIBannerLocation.bottomStart:
      case NUIBannerLocation.bottomEnd:
        return height - _kBottomOffset;
      case NUIBannerLocation.topStart:
      case NUIBannerLocation.topEnd:
        return 0.0;
    }
    return null;
  }

  double get _rotation {
    assert(location != null);
    assert(layoutDirection != null);
    switch (layoutDirection) {
      case TextDirection.rtl:
        switch (location) {
          case NUIBannerLocation.bottomStart:
          case NUIBannerLocation.topEnd:
            return -math.pi / 4.0;
          case NUIBannerLocation.bottomEnd:
          case NUIBannerLocation.topStart:
            return math.pi / 4.0;
        }
        break;
      case TextDirection.ltr:
        switch (location) {
          case NUIBannerLocation.bottomStart:
          case NUIBannerLocation.topEnd:
            return math.pi / 4.0;
          case NUIBannerLocation.bottomEnd:
          case NUIBannerLocation.topStart:
            return -math.pi / 4.0;
        }
        break;
    }
    return null;
  }
}

class NUIBanner extends StatelessWidget {

  const NUIBanner({
    Key key,
    this.child,
    @required this.message,
    this.textDirection,
    @required this.location,
    this.layoutDirection,
    this.color,
    this.textStyle,
    this.gradient,
    this.height: 12,
    this.shadow: const BoxShadow(
      color: Colors.black12,
      blurRadius: 6.0,
    )
  }) : assert(message != null),
        assert(location != null),
        assert(color != null),
        assert(textStyle != null),
        super(key: key);

  final double height;
  final Widget child;
  final String message;
  final TextDirection textDirection;
  final NUIBannerLocation location;
  final TextDirection layoutDirection;
  final LinearGradient gradient;
  final Color color;
  final TextStyle textStyle;
  final BoxShadow shadow;

  @override
  Widget build(BuildContext context) {
    assert((textDirection != null && layoutDirection != null) || debugCheckHasDirectionality(context));
    return CustomPaint(
      foregroundPainter: BannerPainter(
        message: message,
        gradient: gradient,
        height: height,
        textDirection: textDirection ?? Directionality.of(context),
        location: location,
        layoutDirection: layoutDirection ?? Directionality.of(context),
        color: color,
        textStyle: textStyle,
      ),
      child: child,
    );
  }

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties.add(StringProperty('message', message, showName: false));
    properties.add(EnumProperty<TextDirection>('textDirection', textDirection, defaultValue: null));
    properties.add(EnumProperty<NUIBannerLocation>('location', location));
    properties.add(EnumProperty<TextDirection>('layoutDirection', layoutDirection, defaultValue: null));
    properties.add(ColorProperty('color', color, showName: false));
    textStyle?.debugFillProperties(properties, prefix: 'text ');
  }
}
