import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:nui_core/nui_core.dart';
import 'nui_tooltip_overlay.dart';

typedef NUIDropDownButtonBuilder = Widget Function(BuildContext context, bool expanded, NUIDropDownButtonController controller);
typedef NUIItemDropDownButtonBuilder<T> = Widget Function(BuildContext context, T data, NUIDropDownButtonController controller);
typedef NUIDropDownCustomPosition = Offset Function();

class NUIDropDownButtonController{
  Function _dismiss;
  Function _expand;

  void dismiss(){
    if(_dismiss != null) _dismiss();
  }

  void expand(){
    if(_expand != null) _expand();
  }
}

class NUIDropDownButton<T> extends StatefulWidget {
  NUIDropDownButton({@required this.buttonBuilder, @required this.itemBuilder, this.controller, this.animate: true, this.customPosition, this.overlayColor: Colors.transparent, this.dropdownDecoration, this.positionOffset, this.customWidth, this.maxHeight, this.selections, this.overallPadding});

  final NUIDropDownButtonController controller;
  final NUIDropDownButtonBuilder buttonBuilder;
  final EdgeInsets overallPadding;
  final NUIItemDropDownButtonBuilder itemBuilder;
  final BoxDecoration dropdownDecoration;
  final double customWidth;
  final double maxHeight;
  final List<T> selections;
  final Offset positionOffset;
  final Color overlayColor;
  final bool animate;
  final NUIDropDownCustomPosition customPosition;

  @override
  _NUIDropDownButtonState createState() => _NUIDropDownButtonState<T>();
}

class _NUIDropDownButtonState<T> extends State<NUIDropDownButton> with NUITooltipOverlayMixin{
  GlobalKey parentKey;
  NUIDropDownButtonController controller;
  double boxWidth;
  double boxHeight;

  @override
  void initState(){
    super.initState();
    parentKey = GlobalKey();
    boxWidth = 100;
    boxHeight = 50;
    controller = widget.controller ?? NUIDropDownButtonController();

    controller._expand = (){
      _show();
    };

    controller._dismiss = (){
      _cancel();
    };
  }

  void _cancel(){
    if(showingOverlay) cancelTooltipOverlay();
    setState(() {

    });
  }

  void _show(){
    if(!showingOverlay) renderTooltip(context, "default");
    setState(() {

    });
  }

  void _toggleDropdown(){
    if(showingOverlay){
      _cancel();
    }
    else{
      _show();
    }
  }

  @override
  Widget build(BuildContext context) {
    return GestureDetector(
      behavior: HitTestBehavior.translucent,
      onTap: (){
        _toggleDropdown();
      },
      child: AnimatedSwitcher(
        duration: Duration(milliseconds: 250),
        child: LayoutBuilder(
          key: parentKey,
          builder: (context, box){
            boxHeight = box.maxHeight;
            boxWidth = box.maxWidth;
            return Container(
              key: Key(showingOverlay.toString()),
              child: widget.buttonBuilder(context, showingOverlay, controller),
            );
          }
        ),
      ),
    );
  }

  Offset _getPositionOffset(){
    return widget.positionOffset ?? Offset(0, 0);
  }

  EdgeInsets _getOverallPadding(){
    return widget.overallPadding ?? EdgeInsets.all(0);
  }

  BoxConstraints _getBoxConstraints(){
    if(widget.maxHeight != null){
      return BoxConstraints(maxHeight: widget.maxHeight);
    }
    return null;
  }

  BoxDecoration _getBoxDecoration(){
    return widget.dropdownDecoration ?? NUIBoxDecoration(
      color: Colors.white,
      borderRadius: BorderRadius.circular(8),
      shadowRadius: 5,
      shadowOffset: Offset(1, 1),
      shadowColor: Colors.black12
    );
  }

  @override
  Widget getTooltipOverlay(BuildContext context, {Map<String, dynamic> params}) {
    RenderBox box = parentKey.currentContext.findRenderObject();

    //Use custom position if specified, else use default position
    Offset parentPosition = widget.customPosition != null ? widget.customPosition() : box.localToGlobal(Offset.zero);
    final positionOffset = _getPositionOffset();
    Offset position = Offset(parentPosition.dx + positionOffset.dx, parentPosition.dy + positionOffset.dy);

    return _renderDropdownContent(context, position);
  }

  Widget _renderDropdownContent(BuildContext context, Offset position){
    return NUIDropDownButtonContent<T>(
      overlayColor: widget.overlayColor ?? Colors.transparent,
      boxConstraints: _getBoxConstraints(),
      padding: _getOverallPadding(),
      decoration: _getBoxDecoration(),
      position: position,
      width: widget.customWidth ?? boxWidth,
      itemBuilder: widget.itemBuilder,
      controller: controller,
      selections: widget.selections,
      animate: widget.animate,
    );
  }
}

class NUIDropDownButtonContent<T> extends StatefulWidget {
  NUIDropDownButtonContent({this.overlayColor: Colors.transparent, @required this.position, this.animate: true, this.padding, this.decoration, this.boxConstraints, this.width, this.selections, this.itemBuilder, this.controller});
  final Color overlayColor;
  final Offset position;
  final List<T> selections;
  final double width;
  final BoxConstraints boxConstraints;
  final EdgeInsets padding;
  final BoxDecoration decoration;
  final NUIDropDownButtonController controller;
  final NUIItemDropDownButtonBuilder<T> itemBuilder;
  final bool animate;
  @override
  _NUIDropDownButtonContentState createState() => _NUIDropDownButtonContentState();
}

class _NUIDropDownButtonContentState extends State<NUIDropDownButtonContent> {
  bool initiated = false;

  @override
  void initState(){
    super.initState();
    if(widget.animate == true){
      initiated = false;
      NUIAsync.delay(Duration(milliseconds: 200), () {
        setState(() {
          initiated = true;
        });
      });
    }
    else{
      initiated = true;
    }
  }

  @override
  Widget build(BuildContext context) {
    return AnimatedOpacity(
      opacity: initiated ? 1 : 0,
      duration: Duration(milliseconds: 300),
      child: Container(
        width: double.infinity,
        height: double.infinity,
        child: Stack(
          children: [
            GestureDetector(
              behavior: HitTestBehavior.translucent,
              onTap: (){
                widget.controller.dismiss();
              },
              child: Container(
                width: double.infinity,
                height: double.infinity,
                color: widget.overlayColor ?? Colors.transparent,
              ),
            ),
            Positioned(
                top: widget.position.dy,
                left: widget.position.dx,
                child: Container(
                  width: widget.width,
                  constraints: widget.boxConstraints,
                  padding: widget.padding,
                  decoration: widget.decoration,
                  child: SingleChildScrollView(
                    child: Column(
                      children: widget.selections.map((e){
                        return widget.itemBuilder(context, e, widget.controller);
                      }).toList(),
                    ),
                  ),
                )
            ),
          ],
        ),
      ),
    );
  }
}

