import 'package:flutter/material.dart';

typedef NUISpotlightSwitchListener = bool Function();
typedef NUISpotlightIndexSwitchListener = bool Function(int spotlight);
typedef NUISpotlightOverlayBuilder = Widget Function(BuildContext context, BoxConstraints box, Offset position, NUISpotlightController spotlightController);

class NUISpotlightController{
  NUISpotlightSwitchListener _next;
  NUISpotlightSwitchListener _previous;
  NUISpotlightIndexSwitchListener _focus;

  bool focus(int spotlightIndex){
    if(_focus != null) return _focus(spotlightIndex);
    return false;
  }

  bool next(){
    if(_next != null) return _next();
    return false;
  }
  bool previous(){
    if(_previous != null) return _previous();
    return false;
  }
}

class NUISpotlightContainer extends StatefulWidget {
  final Color dimColor;
  final bool animation;
  final bool enabled;
  final Function onTap;
  final bool ignoring;
  final Widget child;
  final int initialSpotlightIndex;
  final NUISpotlightController controller;

  NUISpotlightContainer({
    Key key,
    @required this.child,
    this.controller,
    this.initialSpotlightIndex,
    this.dimColor,
    this.animation = false,
    this.enabled = true,
    this.onTap,
    this.ignoring = true
  }) : super(key: key);

  static NUISpotlightContainerState of(BuildContext context) => context.findAncestorStateOfType<NUISpotlightContainerState>();

  @override
  NUISpotlightContainerState createState() => NUISpotlightContainerState();
}

class NUISpotlightContainerState extends State<NUISpotlightContainer> with SingleTickerProviderStateMixin {
  GlobalKey stickyKey = GlobalKey();
  AnimationController _controller;
  Animation _animation;
  double _fraction = 1.0;
  bool _visible = false;
  int spotlightIndex;
  NUISpotlightController controller;

  Map<int, _NUISpotlightTargetState> spotlights;

  @override
  void initState() {
    super.initState();
    spotlights = Map();
    spotlightIndex = widget.initialSpotlightIndex ?? -1;
    if (widget.animation) {
      _controller = AnimationController(
        duration: Duration(milliseconds: 500),
        vsync: this,
      );

      _animation = Tween(begin: 0.0, end: 1.0)
          .animate(CurvedAnimation(parent: _controller, curve: Curves.ease))
        ..addListener(() {
          setState(() {
            _fraction = _animation.value;
          });
        })
        ..addStatusListener((status) {
          if (status == AnimationStatus.completed) {
            setState(() {
              _visible = true;
            });
          } else if (status == AnimationStatus.dismissed) {
            setState(() {
              _visible = false;
            });
          }
        });

      _controller.forward();
    }
    controller = widget.controller ?? NUISpotlightController();
    controller._previous = (){
      if(spotlightIndex - 1 > 0){
        setState(() {
          spotlightIndex--;
        });
        return true;
      }
      return false;
    };
    controller._next = (){
      if(spotlightIndex < 0){
        setState(() {
          spotlightIndex = 0;
          print("NUISpotlightContainer: Focusing spotlight on $spotlightIndex");
        });
        return true;
      }
      else {
        if (spotlightIndex + 1 < spotlights.length) {
          setState(() {
            spotlightIndex++;
            print("NUISpotlightContainer: Focusing spotlight on $spotlightIndex");
          });
          return true;
        }
        else{
          setState(() {
            spotlightIndex = -1;
            print("NUISpotlightContainer: Completing all spotlights");
          });
        }
      }
      return false;
    };
  }

  void registerField(String name, _NUISpotlightTargetState field) {
    print("NUISpotlightContainer: Registering spotlight item: $name");
    spotlights[field.spotlightIndex] = field;
  }

  void unregisterField(String name, _NUISpotlightTargetState field) {
    print("NUISpotlightContainer: Unregistering spotlight item: $name");
    spotlights.remove(field.spotlightIndex);
  }

  _NUISpotlightTargetState _getCurrentSpotlight(){
    final spotlight = spotlights[spotlightIndex];
    if(spotlight == null){
      print("NUISpotlightContainer: Failed to get spotlight for index $spotlightIndex");
    }
    else{
      print("NUISpotlightContainer: Spotlight found for index $spotlightIndex");
    }
    return spotlight;
  }

  @override
  void dispose() {
    _controller?.dispose();
    super.dispose();
  }

  @override
  void didUpdateWidget(NUISpotlightContainer oldWidget) {
    _controller?.reset();
    _controller?.forward();
    super.didUpdateWidget(oldWidget);
  }

  @override
  Widget build(BuildContext context) {
    final currentSpotlight = _getCurrentSpotlight();
    return Stack(
      key: stickyKey,
      children: widget.enabled ?
      [
        // Child does not respond to gestures while displaying spotlight
        IgnorePointer(
            ignoring: widget.ignoring && spotlightIndex > -1,
            child: widget.child
        ),
        GestureDetector(
          child: AnimatedSwitcher(
            duration: Duration(milliseconds: 300),
            child: spotlightIndex > -1 && currentSpotlight != null ? ClipPath(
              key: Key(spotlightIndex?.toString()),
              clipper: currentSpotlight.getClipper(),
              child: Stack(
                children: <Widget>[
                  Container(
                    width: double.infinity,
                    height: double.infinity,
                    color: widget.dimColor ?? Colors.black12,
                  ),
                  Opacity(
                    opacity: _visible ? 1.0 : 0.0,
                    child: currentSpotlight._getOverlay(context) ?? SizedBox.shrink(),
                  ),
                ],
              ),
            ) : SizedBox.shrink(),
          ),
          onTap: () {
            if (widget.onTap != null) {
              widget.onTap();
            }
          },
        ),
      ] : [
        widget.child,
      ],
    );
  }

  Offset adjustCenter(Offset offset) {
    // Get the translation vector (widget position) from NUISpotlightContainer
    var object = stickyKey.currentContext?.findRenderObject();
    var translation = object?.getTransformTo(null)?.getTranslation();
    if (translation != null) {
      // Adjust center
      double dx = offset.dx - translation.x;
      double dy = offset.dy - translation.y;
      return Offset(dx, dy);
    } else {
      return offset;
    }
  }
}


class NUISpotlightTarget extends StatefulWidget {
  NUISpotlightTarget({@required this.name, @required this.spotlightIndex, this.spotlightMargin, this.customSize, @required this.child, this.borderRadius, this.overlayBuilder});

  final String name;
  final int spotlightIndex;
  final Widget child;
  final BorderRadius borderRadius;
  final Offset spotlightMargin;
  final Size customSize;
  final NUISpotlightOverlayBuilder overlayBuilder;

  @override
  _NUISpotlightTargetState createState() => _NUISpotlightTargetState();
}

class _NUISpotlightTargetState extends State<NUISpotlightTarget> {
  NUISpotlightContainerState _containerState;
  int spotlightIndex;
  String name;
  BorderRadius borderRadius;
  GlobalKey layoutKey;
  BoxConstraints box;
  Offset spotlightMargin;
  NUISpotlightOverlayBuilder overlayBuilder;
  Size customSize;

  @override
  void initState(){
    super.initState();
    layoutKey = GlobalKey();
    customSize = widget.customSize;
    spotlightIndex = widget.spotlightIndex;
    name = widget.name;
    borderRadius = widget.borderRadius;
    overlayBuilder = widget.overlayBuilder;
    spotlightMargin = widget.spotlightMargin;

    _containerState = NUISpotlightContainer.of(context);
    if(_containerState != null) {
      _containerState?.registerField(name, this);
    }
    else{
      print("NUISpotlightContainer; Failed to get parent spotlight container");
    }
  }

  Widget _getOverlay(BuildContext context){
    if(overlayBuilder == null) return null;
    return overlayBuilder(context, box, _getPosition(), _containerState?.controller);
  }

  @override
  Widget build(BuildContext context) {
    return Container(
      key: layoutKey,
      child: LayoutBuilder(
        builder: (context, box){
          this.box = box;
          return widget.child;
        },
      ),
    );
  }

  Offset _getPosition(){
    RenderBox renderBox = layoutKey.currentContext.findRenderObject();
    Offset position = renderBox.localToGlobal(Offset.zero);
    print("NUISpotlightContainer: Spotlight of $spotlightIndex is at position $position");
    Offset processPosition = Offset(position.dx + (spotlightMargin?.dx ?? 0), position.dy + (spotlightMargin?.dy ?? 0));
    return processPosition;
  }

  CustomClipper<Path> getClipper(){
    Offset position = _getPosition();
    RenderBox renderBox = layoutKey.currentContext.findRenderObject();

    final width = customSize?.width ?? renderBox.size.width;
    final height = customSize?.height ?? renderBox.size.height;
    return NUIInvertedRectSpotlightClipper(position, width, height, borderRadius);
  }

  @override
  void dispose(){
    _containerState?.unregisterField(name, this);
    super.dispose();
  }
}


class NUIInvertedCircleSpotlightClipper extends CustomClipper<Path> {
  final Offset center;
  final double radius;

  NUIInvertedCircleSpotlightClipper(this.center, this.radius);

  @override
  Path getClip(Size size) {
    return Path()
      ..addOval(
        Rect.fromCircle(
          center: center ?? Offset(size.width / 2, size.height / 2),
          radius: radius ?? size.width * 0.25,
        ),
      )
    // Overall drawing (全体描画)
      ..addRect(Rect.fromLTWH(0.0, 0.0, size.width, size.height))
      ..fillType = PathFillType.evenOdd;
  }

  @override
  bool shouldReclip(NUIInvertedCircleSpotlightClipper oldClipper) {
    return oldClipper.center != center || oldClipper.radius != radius;
  }
}

class NUIInvertedRectSpotlightClipper extends CustomClipper<Path> {
  final Offset position;
  final double width;
  final double height;
  final BorderRadius borderRadius;

  NUIInvertedRectSpotlightClipper(this.position, this.width, this.height, this.borderRadius);

  @override
  Path getClip(Size size) {
    final path = Path()
      ..addPath(
          Path()
            ..lineTo(0.0, height - borderRadius.bottomLeft.y)
            ..arcToPoint(Offset(borderRadius.bottomLeft.x, height), radius: Radius.circular(borderRadius.bottomLeft.x), rotation: 180, clockwise: false)
            ..lineTo(width - borderRadius.bottomRight.x, height)
            ..arcToPoint(Offset(width, height - borderRadius.bottomRight.y), radius: Radius.circular(borderRadius.bottomRight.x), rotation: 180, clockwise: false)
            ..lineTo(width, borderRadius.topRight.y)
            ..arcToPoint(Offset(width - borderRadius.topRight.x, 0), radius: Radius.circular(borderRadius.topRight.x), rotation: 180, clockwise: false)
            ..lineTo(borderRadius.topLeft.x, 0)
            ..arcToPoint(Offset(0, borderRadius.topLeft.y), radius: Radius.circular(borderRadius.topLeft.x), rotation: 180, clockwise: false)
          ,
          position
      )
      ..addRect(Rect.fromLTWH(0.0, 0.0, size.width, size.height))
      ..fillType = PathFillType.evenOdd;

    return path;
  }

  @override
  bool shouldReclip(NUIInvertedRectSpotlightClipper oldClipper) {
    return oldClipper.position != position || oldClipper.width != width || oldClipper.height != height || oldClipper.borderRadius != borderRadius;
  }
}

class NUISpotlightContentPointerShape extends CustomClipper<Path> {
  bool pointUp;

  NUISpotlightContentPointerShape({this.pointUp: false});

  @override
  Path getClip(Size size) {
    var path = Path();

    if(pointUp) {
      path.moveTo(size.width / 2, 0);
      path.lineTo(0, size.height);
      path.lineTo(size.width, size.height);
      path.close();
    }
    else{
      path.moveTo(size.width / 2, size.height);
      path.lineTo(0, 0);
      path.lineTo(size.width, 0);
      path.close();
    }

    return path;
  }

  @override
  bool shouldRepaint(CustomPainter oldDelegate) {
    return false;
  }

  @override
  bool shouldReclip(covariant CustomClipper<Path> oldClipper) {
    return false;
  }
}